@@ -49,6 +49,7 @@
 #include "sj-structures.h"
 #include "sj-error.h"
 #include "sj-util.h"
+#include "sj-main.h"
 #include "sj-prefs.h"
 #include "sj-play.h"
 #include "bacon-volume.h"
@@ -136,7 +137,7 @@ sj_stock_init (void)
   static gboolean initialized = FALSE;
   static GtkIconFactory *sj_icon_factory = NULL;
 
-  static GtkStockItem sj_stock_items[] =
+  static const GtkStockItem sj_stock_items[] =
   {
     { SJ_STOCK_PLAYING, NULL, 0, 0, NULL },
     { SJ_STOCK_RECORDING, NULL, 0, 0, NULL },
@@ -782,6 +783,9 @@ media_removed_cb (NautilusBurnDrive *drive,
     /* FIXME: recover? */
   }
 
+  /* first make sure we're not playing */
+  stop_playback ();
+
   sj_debug (DEBUG_CD, "Media removed from device %s\n", nautilus_burn_drive_get_device (drive));
   stop_ui_hack ();
   update_ui_for_album (NULL);
@@ -1218,7 +1222,7 @@ void on_contents_activate(GtkWidget *button, gpointer user_data) {
 }
 
 
-static const char* genres[] = {
+static const char* const genres[] = {
   "Ambient",
   "Blues",
   "Classical",
@@ -1240,7 +1244,7 @@ static const char* genres[] = {
 
 static GtkTreeModel* create_genre_list(void) {
   GtkListStore *store;
-  const char **g = genres;
+  const char * const *g = genres;
 
   store = gtk_list_store_new (1, G_TYPE_STRING);
 
